import dataclasses
import re
import typing
import logging
import deepdog.indexify
import pathlib
import csv

_logger = logging.getLogger(__name__)

FILENAME_REGEX = r"(?P<timestamp>\d{8}-\d{6})-(?P<filename_slug>.*)\.realdata\.fast_filter\.bayesrun\.csv"

MODEL_REGEXES = [
	r"geom_(?P<xmin>-?\d+)_(?P<xmax>-?\d+)_(?P<ymin>-?\d+)_(?P<ymax>-?\d+)_(?P<zmin>-?\d+)_(?P<zmax>-?\d+)-orientation_(?P<orientation>free|fixedxy|fixedz)-dipole_count_(?P<avg_filled>\d+)_(?P<field_name>\w*)"
]

FILE_SLUG_REGEXES = [
	r"mock_tarucha-(?P<job_index>\d+)",
	r"(?:(?P<mock>mock)_)?tarucha(?:_(?P<tarucha_run_id>\d+))?-(?P<job_index>\d+)",
]


@dataclasses.dataclass
class BayesrunOutputFilename:
	timestamp: str
	filename_slug: str
	path: pathlib.Path


@dataclasses.dataclass
class BayesrunColumnParsed:
	"""
	class for parsing a bayesrun while pulling certain special fields out
	"""

	def __init__(self, groupdict: typing.Dict[str, str]):
		self.column_field = groupdict["field_name"]
		self.model_field_dict = {
			k: v for k, v in groupdict.items() if k != "field_name"
		}

	def __str__(self):
		return f"BayesrunColumnParsed[{self.column_field}: {self.model_field_dict}]"


@dataclasses.dataclass
class BayesrunModelResult:
	parsed_model_keys: typing.Dict[str, str]
	success: int
	count: int


@dataclasses.dataclass
class BayesrunOutput:
	filename: BayesrunOutputFilename
	data: typing.Dict["str", typing.Any]
	results: typing.Sequence[BayesrunModelResult]


def _batch_iterable_into_chunks(iterable, n=1):
	"""
	utility for batching bayesrun files where columns appear in threes
	"""
	for ndx in range(0, len(iterable), n):
		yield iterable[ndx : min(ndx + n, len(iterable))]


def _parse_bayesrun_column(
	column: str,
) -> typing.Optional[BayesrunColumnParsed]:
	"""
	Tries one by one all of a predefined list of regexes that I might have used in the past.
	Returns the groupdict for the first match, or None if no match found.
	"""
	for pattern in MODEL_REGEXES:
		match = re.match(pattern, column)
		if match:
			return BayesrunColumnParsed(match.groupdict())
	else:
		return None


def _parse_bayesrun_row(
	row: typing.Dict[str, str],
) -> typing.Sequence[BayesrunModelResult]:

	results = []
	batched_keys = _batch_iterable_into_chunks(list(row.keys()), 3)
	for model_keys in batched_keys:
		parsed = [_parse_bayesrun_column(column) for column in model_keys]
		values = [row[column] for column in model_keys]
		if parsed[0] is None:
			raise ValueError(f"no viable success row found for keys {model_keys}")
		if parsed[1] is None:
			raise ValueError(f"no viable count row found for keys {model_keys}")
		if parsed[0].column_field != "success":
			raise ValueError(f"The column {model_keys[0]} is not a success field")
		if parsed[1].column_field != "count":
			raise ValueError(f"The column {model_keys[1]} is not a count field")
		parsed_keys = parsed[0].model_field_dict
		success = int(values[0])
		count = int(values[1])
		results.append(
			BayesrunModelResult(
				parsed_model_keys=parsed_keys,
				success=success,
				count=count,
			)
		)
	return results


def _parse_output_filename(file: pathlib.Path) -> BayesrunOutputFilename:
	filename = file.name
	match = re.match(FILENAME_REGEX, filename)
	if not match:
		raise ValueError(f"{filename} was not a valid bayesrun output")
	groups = match.groupdict()
	return BayesrunOutputFilename(
		timestamp=groups["timestamp"], filename_slug=groups["filename_slug"], path=file
	)


def _parse_file_slug(slug: str) -> typing.Optional[typing.Dict[str, str]]:
	for pattern in FILE_SLUG_REGEXES:
		match = re.match(pattern, slug)
		if match:
			return match.groupdict()
	else:
		return None


def read_output_file(
	file: pathlib.Path, indexifier: typing.Optional[deepdog.indexify.Indexifier]
) -> BayesrunOutput:

	parsed_filename = tag = _parse_output_filename(file)
	out = BayesrunOutput(filename=parsed_filename, data={}, results=[])

	out.data.update(dataclasses.asdict(tag))
	parsed_tag = _parse_file_slug(parsed_filename.filename_slug)
	if parsed_tag is None:
		_logger.warning(
			f"Could not parse {tag} against any matching regexes. Going to skip tag parsing"
		)
	else:
		out.data.update(parsed_tag)
		if indexifier is not None:
			try:
				job_index = parsed_tag["job_index"]
				indexified = indexifier.indexify(int(job_index))
				out.data.update(indexified)
			except KeyError:
				# This isn't really that important of an error, apart from the warning
				_logger.warning(
					f"Parsed tag to {parsed_tag}, and attempted to indexify but no job_index key was found. skipping and moving on"
				)

	with file.open() as input_file:
		reader = csv.DictReader(input_file)
		rows = [r for r in reader]
		if len(rows) == 1:
			row = rows[0]
		else:
			raise ValueError(f"Confused about having multiple rows in {file.name}")
	results = _parse_bayesrun_row(row)

	out.results = results

	return out



from typing import Sequence
from deepdog.direct_monte_carlo.direct_mc import DirectMonteCarloFilter
import numpy


class ComposedDMCFilter(DirectMonteCarloFilter):
	def __init__(self, filters: Sequence[DirectMonteCarloFilter]):
		self.filters = filters

	def filter_samples(self, samples: numpy.ndarray) -> numpy.ndarray:
		current_sample = samples
		for filter in self.filters:
			current_sample = filter.filter_samples(current_sample)
		return current_sample



import typing
from deepdog.results import BayesrunOutput
import logging
import csv
import tqdm

_logger = logging.getLogger(__name__)


def build_model_dict(
	bayes_outputs: typing.Sequence[BayesrunOutput],
) -> typing.Dict[
	typing.Tuple, typing.Dict[typing.Tuple, typing.Dict["str", typing.Any]]
]:
	"""
	Maybe someday do something smarter with the coalescing and stuff but don't want to so i won't
	"""
	# assume that everything is well formatted and the keys are the same across entire list and initialise list of keys.
	# model dict will contain a model_key: {calculation_dict} where each calculation_dict represents a single calculation for that model,
	# the uncoalesced version, keyed by the specific file keys
	model_dict: typing.Dict[
		typing.Tuple, typing.Dict[typing.Tuple, typing.Dict["str", typing.Any]]
	] = {}

	_logger.info("building model dict")
	for out in tqdm.tqdm(bayes_outputs, desc="reading outputs", leave=False):
		for model_result in out.results:
			model_key = tuple(v for v in model_result.parsed_model_keys.values())
			if model_key not in model_dict:
				model_dict[model_key] = {}
			calculation_dict = model_dict[model_key]
			calculation_key = tuple(v for v in out.data.values())
			if calculation_key not in calculation_dict:
				calculation_dict[calculation_key] = {
					"_model_key_dict": model_result.parsed_model_keys,
					"_calculation_key_dict": out.data,
					"success": model_result.success,
					"count": model_result.count,
				}
			else:
				raise ValueError(
					f"Got {calculation_key} twice for model_key {model_key}"
				)

	return model_dict


def write_uncoalesced_dict(
	uncoalesced_output_filename: typing.Optional[str],
	uncoalesced_model_dict: typing.Dict[
		typing.Tuple, typing.Dict[typing.Tuple, typing.Dict["str", typing.Any]]
	],
):
	if uncoalesced_output_filename is None or uncoalesced_output_filename == "":
		_logger.warning("Not provided a uncoalesced filename, not going to try")
		return

	first_value = next(iter(next(iter(uncoalesced_model_dict.values())).values()))
	model_field_names = set(first_value["_model_key_dict"].keys())
	calculation_field_names = set(first_value["_calculation_key_dict"].keys())
	if not (set(model_field_names).isdisjoint(calculation_field_names)):
		_logger.info(f"Detected model field names {model_field_names}")
		_logger.info(f"Detected calculation field names {calculation_field_names}")
		_logger.warning(
			f"model field names {model_field_names} and calculation {calculation_field_names} have an overlap, which is possibly a problem"
		)
	collected_fieldnames = list(model_field_names)
	collected_fieldnames.extend(calculation_field_names)
	collected_fieldnames.extend(["success", "count"])
	_logger.info(f"Full uncoalesced fieldnames are {collected_fieldnames}")
	with open(uncoalesced_output_filename, "w", newline="") as uncoalesced_output_file:
		writer = csv.DictWriter(
			uncoalesced_output_file, fieldnames=collected_fieldnames
		)
		writer.writeheader()

		for model_dict in uncoalesced_model_dict.values():
			for calculation in model_dict.values():
				row = calculation["_model_key_dict"].copy()
				row.update(calculation["_calculation_key_dict"].copy())
				row.update(
					{
						"success": calculation["success"],
						"count": calculation["count"],
					}
				)
				writer.writerow(row)


def coalesced_dict(
	uncoalesced_model_dict: typing.Dict[
		typing.Tuple, typing.Dict[typing.Tuple, typing.Dict["str", typing.Any]]
	],
	minimum_count: float = 0.1,
):
	"""
	pass in uncoalesced dict
	the minimum_count field is what we use to make sure our probs are never zero
	"""
	coalesced_dict = {}

	# we are already iterating so for no reason because performance really doesn't matter let's count the keys ourselves
	num_keys = 0

	# first pass coalesce
	for model_key, model_dict in uncoalesced_model_dict.items():
		num_keys += 1
		for calculation in model_dict.values():
			if model_key not in coalesced_dict:
				coalesced_dict[model_key] = {
					"_model_key_dict": calculation["_model_key_dict"].copy(),
					"calculations_coalesced": 0,
					"count": 0,
					"success": 0,
				}
			sub_dict = coalesced_dict[model_key]
			sub_dict["calculations_coalesced"] += 1
			sub_dict["count"] += calculation["count"]
			sub_dict["success"] += calculation["success"]

	# second pass do probability calculation

	prior = 1 / num_keys
	_logger.info(f"Got {num_keys} model keys, so our prior will be {prior}")

	total_weight = 0
	for coalesced_model_dict in coalesced_dict.values():
		model_weight = (
			max(minimum_count, coalesced_model_dict["success"])
			/ coalesced_model_dict["count"]
		) * prior
		total_weight += model_weight

	total_prob = 0
	for coalesced_model_dict in coalesced_dict.values():
		model_weight = (
			max(minimum_count, coalesced_model_dict["success"])
			/ coalesced_model_dict["count"]
		)
		prob = model_weight * prior / total_weight
		coalesced_model_dict["prob"] = prob
		total_prob += prob

	_logger.debug(
		f"Got a total probability of {total_prob}, which should be close to 1 up to float/rounding error"
	)
	return coalesced_dict


def write_coalesced_dict(
	coalesced_output_filename: typing.Optional[str],
	coalesced_model_dict: typing.Dict[typing.Tuple, typing.Dict["str", typing.Any]],
):
	if coalesced_output_filename is None or coalesced_output_filename == "":
		_logger.warning("Not provided a uncoalesced filename, not going to try")
		return

	first_value = next(iter(coalesced_model_dict.values()))
	model_field_names = set(first_value["_model_key_dict"].keys())
	_logger.info(f"Detected model field names {model_field_names}")

	collected_fieldnames = list(model_field_names)
	collected_fieldnames.extend(["calculations_coalesced", "success", "count", "prob"])
	with open(coalesced_output_filename, "w", newline="") as coalesced_output_file:
		writer = csv.DictWriter(coalesced_output_file, fieldnames=collected_fieldnames)
		writer.writeheader()

		for model_dict in coalesced_model_dict.values():
			row = model_dict["_model_key_dict"].copy()
			row.update(
				{
					"calculations_coalesced": model_dict["calculations_coalesced"],
					"success": model_dict["success"],
					"count": model_dict["count"],
					"prob": model_dict["prob"],
				}
			)
			writer.writerow(row)



import logging
import argparse
import json
import deepdog.cli.probs.args
import deepdog.cli.probs.dicts
import deepdog.results
import deepdog.indexify
import pathlib
import tqdm
import tqdm.contrib.logging


_logger = logging.getLogger(__name__)


def set_up_logging(log_file: str):

	log_pattern = "%(asctime)s | %(levelname)-7s | %(name)s:%(lineno)d | %(message)s"
	if log_file is None:
		handlers = [
			logging.StreamHandler(),
		]
	else:
		handlers = [logging.StreamHandler(), logging.FileHandler(log_file)]
	logging.basicConfig(
		level=logging.DEBUG,
		format=log_pattern,
		# it's okay to ignore this mypy error because who cares about logger handler types
		handlers=handlers,  # type: ignore
	)
	logging.captureWarnings(True)


def main(args: argparse.Namespace):
	"""
	Main function with passed in arguments and no additional logging setup in case we want to extract out later
	"""

	with tqdm.contrib.logging.logging_redirect_tqdm():
		_logger.info(f"args: {args}")

		try:
			if args.coalesced_keys:
				raise NotImplementedError(
					"Currently not supporting coalesced keys, but maybe in future"
				)
		except AttributeError:
			# we don't care if this is missing because we don't actually want it to be there
			pass

		indexifier = None
		if args.indexify_json:
			with open(args.indexify_json, "r") as indexify_json_file:
				indexify_spec = json.load(indexify_json_file)
				indexify_data = indexify_spec["indexes"]
				if "seed_spec" in indexify_spec:
					seed_spec = indexify_spec["seed_spec"]
					indexify_data[seed_spec["field_name"]] = list(
						range(seed_spec["num_seeds"])
					)
				# _logger.debug(f"Indexifier data looks like {indexify_data}")
				indexifier = deepdog.indexify.Indexifier(indexify_data)

		bayes_dir = pathlib.Path(args.bayesrun_directory)
		out_files = [f for f in bayes_dir.iterdir() if f.name.endswith("bayesrun.csv")]
		_logger.info(
			f"Reading {len(out_files)} bayesrun.csv files in directory {args.bayesrun_directory}"
		)
		# _logger.info(out_files)
		parsed_output_files = [
			deepdog.results.read_output_file(f, indexifier)
			for f in tqdm.tqdm(out_files, desc="reading files", leave=False)
		]

		_logger.info("building uncoalesced dict")
		uncoalesced_dict = deepdog.cli.probs.dicts.build_model_dict(parsed_output_files)

		if "uncoalesced_outfile" in args and args.uncoalesced_outfile:
			deepdog.cli.probs.dicts.write_uncoalesced_dict(
				args.uncoalesced_outfile, uncoalesced_dict
			)
		else:
			_logger.info("Skipping writing uncoalesced")

		_logger.info("building coalesced dict")
		coalesced = deepdog.cli.probs.dicts.coalesced_dict(uncoalesced_dict)

		if "coalesced_outfile" in args and args.coalesced_outfile:
			deepdog.cli.probs.dicts.write_coalesced_dict(
				args.coalesced_outfile, coalesced
			)
		else:
			_logger.info("Skipping writing coalesced")


def wrapped_main():
	args = deepdog.cli.probs.args.parse_args()
	set_up_logging(args.log_file)
	main(args)



from numpy import ndarray
from deepdog.direct_monte_carlo.direct_mc import DirectMonteCarloFilter
from typing import Sequence
import pdme.measurement
import pdme.measurement.input_types
import pdme.util.fast_nonlocal_spectrum
import pdme.util.fast_v_calc
import numpy


class SingleDotPotentialFilter(DirectMonteCarloFilter):
	def __init__(self, measurements: Sequence[pdme.measurement.DotRangeMeasurement]):
		self.measurements = measurements
		self.dot_inputs = [(measure.r, measure.f) for measure in self.measurements]

		self.dot_inputs_array = pdme.measurement.input_types.dot_inputs_to_array(
			self.dot_inputs
		)
		(
			self.lows,
			self.highs,
		) = pdme.measurement.input_types.dot_range_measurements_low_high_arrays(
			self.measurements
		)

	def filter_samples(self, samples: ndarray) -> ndarray:
		current_sample = samples
		for di, low, high in zip(self.dot_inputs_array, self.lows, self.highs):

			if len(current_sample) < 1:
				break
			vals = pdme.util.fast_v_calc.fast_vs_for_dipoleses(
				numpy.array([di]), current_sample
			)

			current_sample = current_sample[
				numpy.all((vals > low) & (vals < high), axis=1)
			]
		return current_sample


class SingleDotSpinQubitFrequencyFilter(DirectMonteCarloFilter):
	def __init__(self, measurements: Sequence[pdme.measurement.DotRangeMeasurement]):
		self.measurements = measurements
		self.dot_inputs = [(measure.r, measure.f) for measure in self.measurements]

		self.dot_inputs_array = pdme.measurement.input_types.dot_inputs_to_array(
			self.dot_inputs
		)
		(
			self.lows,
			self.highs,
		) = pdme.measurement.input_types.dot_range_measurements_low_high_arrays(
			self.measurements
		)

	# oh no not this again
	def fast_s_spin_qubit_tarucha_apsd_dipoleses(
		self, dot_inputs: numpy.ndarray, dipoleses: numpy.ndarray
	) -> numpy.ndarray:
		"""
		No error correction here baby.
		"""

		# We're going to annotate the indices on this class.
		# Let's define some indices:
		# A -> index of dipoleses configurations
		# j -> within a particular configuration, indexes dipole j
		# measurement_index -> if we have 100 frequencies for example, indexes which one of them it is
		# If we need to use numbers, let's use A -> 2, j -> 10, measurement_index -> 9 for consistency with
		# my other notes

		# axes are [dipole_config_idx A, dipole_idx j, {px, py, pz}3]
		ps = dipoleses[:, :, 0:3]
		# axes are [dipole_config_idx A, dipole_idx j, {sx, sy, sz}3]
		ss = dipoleses[:, :, 3:6]
		# axes are [dipole_config_idx A, dipole_idx j, w], last axis is just 1
		ws = dipoleses[:, :, 6]

		# dot_index is either 0 or 1 for dot1 or dot2
		# hopefully this adhoc grammar is making sense, with the explicit labelling of the values of the last axis in cartesian space
		# axes are [measurement_idx, {dot_index}, {rx, ry, rz}] where the inner {dot_index} is gone
		# [measurement_idx, cartesian3]
		rs = dot_inputs[:, 0:3]
		# axes are [measurement_idx]
		fs = dot_inputs[:, 3]

		# first operation!
		# r1s has shape [measurement_idx, rxs]
		# None inserts an extra axis so the r1s[:, None] has shape
		# [measurement_idx, 1]([rxs]) with the last rxs hidden
		#
		# ss has shape [ A, j, {sx, sy, sz}3], so second term has shape [A, 1, j]([sxs])
		# these broadcast from right to left
		# [	 measurement_idx, 1, rxs]
		# [A,	  1,			   j, sxs]
		# resulting in [A, measurement_idx, j, cart3] sxs rxs are both cart3
		diffses = rs[:, None] - ss[:, None, :]

		# norms takes out axis 3, the last one, giving [A, measurement_idx, j]
		norms = numpy.linalg.norm(diffses, axis=3)

		# _logger.info(f"norms1: {norms1}")
		# _logger.info(f"norms1 shape: {norms1.shape}")
		#
		# diffses1 (A, measurement_idx, j, xs)
		# ps:  (A, j, px)
		# result is (A, measurement_idx, j)
		# intermediate_dot_prod = numpy.einsum("abcd,acd->abc", diffses1, ps)
		# _logger.info(f"dot product shape: {intermediate_dot_prod.shape}")

		# transpose makes it (j, measurement_idx, A)
		# transp_intermediate_dot_prod = numpy.transpose(numpy.einsum("abcd,acd->abc", diffses1, ps) / (norms1**3))

		# transpose of diffses has shape (xs, j, measurement_idx, A)
		# numpy.transpose(diffses1)
		# _logger.info(f"dot product shape: {transp_intermediate_dot_prod.shape}")

		# inner transpose is (j, measurement_idx, A) * (xs, j, measurement_idx, A)
		# next transpose puts it back to (A, measurement_idx, j, xs)
		# p_dot_r_times_r_term = 3 * numpy.transpose(numpy.transpose(numpy.einsum("abcd,acd->abc", diffses1, ps) / (norms1**3)) * numpy.transpose(diffses1))
		# _logger.info(f"p_dot_r_times_r_term: {p_dot_r_times_r_term.shape}")

		# only x axis puts us at (A, measurement_idx, j)
		# p_dot_r_times_r_term_x_only = p_dot_r_times_r_term[:, :, :, 0]
		# _logger.info(f"p_dot_r_times_r_term_x_only.shape: {p_dot_r_times_r_term_x_only.shape}")

		# now to complete the numerator we subtract the ps, which are (A, j, px):
		# slicing off the end gives us (A, j), so we newaxis to get (A, 1, j)
		# _logger.info(ps[:, numpy.newaxis, :, 0].shape)
		alphses = (
			(
				3
				* numpy.transpose(
					numpy.transpose(
						numpy.einsum("abcd,acd->abc", diffses, ps) / (norms**2)
					)
					* numpy.transpose(diffses)
				)[:, :, :, 0]
			)
			- ps[:, numpy.newaxis, :, 0]
		) / (norms**3)

		bses = (
			2
			* numpy.pi
			* ws[:, None, :]
			/ ((2 * numpy.pi * fs[:, None]) ** 2 + 4 * ws[:, None, :] ** 2)
		)

		return numpy.einsum("...j->...", alphses * alphses * bses)

	def filter_samples(self, samples: ndarray) -> ndarray:
		current_sample = samples
		for di, low, high in zip(self.dot_inputs_array, self.lows, self.highs):

			if len(current_sample) < 1:
				break
			vals = self.fast_s_spin_qubit_tarucha_apsd_dipoleses(
				numpy.array([di]), current_sample
			)
			# _logger.info(vals)

			current_sample = current_sample[
				numpy.all((vals > low) & (vals < high), axis=1)
			]
		# _logger.info(f"leaving with {len(current_sample)}")
		return current_sample


class DoubleDotSpinQubitFrequencyFilter(DirectMonteCarloFilter):
	def __init__(
		self,
		pair_phase_measurements: Sequence[pdme.measurement.DotPairRangeMeasurement],
	):
		self.pair_phase_measurements = pair_phase_measurements
		self.dot_pair_inputs = [
			(measure.r1, measure.r2, measure.f)
			for measure in self.pair_phase_measurements
		]
		self.dot_pair_inputs_array = (
			pdme.measurement.input_types.dot_pair_inputs_to_array(self.dot_pair_inputs)
		)
		(
			self.pair_phase_lows,
			self.pair_phase_highs,
		) = pdme.measurement.input_types.dot_range_measurements_low_high_arrays(
			self.pair_phase_measurements
		)

	def filter_samples(self, samples: ndarray) -> ndarray:
		current_sample = samples

		for pi, plow, phigh in zip(
			self.dot_pair_inputs_array, self.pair_phase_lows, self.pair_phase_highs
		):
			if len(current_sample) < 1:
				break

			vals = pdme.util.fast_nonlocal_spectrum.signarg(
				pdme.util.fast_nonlocal_spectrum.fast_s_spin_qubit_tarucha_nonlocal_dipoleses(
					numpy.array([pi]), current_sample
				)
			)
			current_sample = current_sample[
				numpy.all(
					((vals > plow) & (vals < phigh)) | ((vals < plow) & (vals > phigh)),
					axis=1,
				)
			]
		return current_sample



import argparse
import os


def parse_args() -> argparse.Namespace:
	def dir_path(path):
		if os.path.isdir(path):
			return path
		else:
			raise argparse.ArgumentTypeError(f"readable_dir:{path} is not a valid path")

	parser = argparse.ArgumentParser(
		"probs", description="Calculating probability from finished bayesrun"
	)
	parser.add_argument(
		"--log_file",
		type=str,
		help="A filename for logging to, if not provided will only log to stderr",
		default=None,
	)
	parser.add_argument(
		"--bayesrun-directory",
		"-d",
		type=dir_path,
		help="The directory to search for bayesrun files, defaulting to cwd if not passed",
		default=".",
	)
	parser.add_argument(
		"--indexify-json",
		help="A json file with the indexify config for parsing job indexes. Will skip if not present",
		default="",
	)
	parser.add_argument(
		"--coalesced-keys",
		type=str,
		help="A comma separated list of strings over which to coalesce data. By default coalesce over all fields within model names, ignore file level names",
		default="",
	)
	parser.add_argument(
		"--uncoalesced-outfile",
		type=str,
		help="output filename for uncoalesced data. If not provided, will not be written",
		default=None,
	)
	parser.add_argument(
		"--coalesced-outfile",
		type=str,
		help="output filename for coalesced data. If not provided, will not be written",
		default=None,
	)
	return parser.parse_args()



"""
Probably should just include a way to handle the indexify function I reuse so much.

All about breaking an integer into a tuple of values from lists, which is useful because of how we do CHTC runs.
"""
import itertools
import typing
import logging
import math

_logger = logging.getLogger(__name__)


# from https://stackoverflow.com/questions/5228158/cartesian-product-of-a-dictionary-of-lists
def _dict_product(dicts):
	"""
	>>> list(dict_product(dict(number=[1,2], character='ab')))
	[{'character': 'a', 'number': 1},
	{'character': 'a', 'number': 2},
	{'character': 'b', 'number': 1},
	{'character': 'b', 'number': 2}]
	"""
	return list(dict(zip(dicts.keys(), x)) for x in itertools.product(*dicts.values()))


class Indexifier:
	"""
	The order of keys is very important, but collections.OrderedDict is no longer needed in python 3.7.
	I think it's okay to rely on that.
	"""

	def __init__(self, list_dict: typing.Dict[str, typing.Sequence]):
		self.dict = list_dict

	def indexify(self, n: int) -> typing.Dict[str, typing.Any]:
		product_dict = _dict_product(self.dict)
		return product_dict[n]

	def _indexify_indices(self, n: int) -> typing.Sequence[int]:
		"""
		legacy indexify from old scripts, copypast.
		could be used like
		>>> ret = {}
		>>> for k, i in zip(self.dict.keys(), self._indexify_indices):
		>>>		ret[k] = self.dict[k][i]
		>>> return ret
		"""
		weights = [len(v) for v in self.dict.values()]
		N = math.prod(weights)
		curr_n = n
		curr_N = N
		out = []
		for w in weights[:-1]:
			# print(f"current: {curr_N}, {curr_n}, {curr_n // w}")
			curr_N = curr_N // w  # should be int division anyway
			out.append(curr_n // curr_N)
			curr_n = curr_n % curr_N
		return out



import csv
import pdme.model
import pdme.measurement
import pdme.measurement.input_types
import pdme.subspace_simulation
import datetime
from typing import Tuple, Dict, NewType, Any, Sequence
from dataclasses import dataclass
import logging
import numpy
import numpy.random
import pdme.util.fast_v_calc
import multiprocessing

_logger = logging.getLogger(__name__)


@dataclass
class DirectMonteCarloResult:
	successes: int
	monte_carlo_count: int
	likelihood: float
	model_name: str


@dataclass
class DirectMonteCarloConfig:
	monte_carlo_count_per_cycle: int = 10000
	monte_carlo_cycles: int = 10
	target_success: int = 100
	max_monte_carlo_cycles_steps: int = 10
	monte_carlo_seed: int = 1234
	write_successes_to_file: bool = False
	tag: str = ""
	cap_core_count: int = 0  # 0 means cap at num cores - 1
	chunk_size: int = 50
	write_bayesrun_file = True
	# chunk size of some kind


# Aliasing dict as a generic data container
DirectMonteCarloData = NewType("DirectMonteCarloData", Dict[str, Any])


class DirectMonteCarloFilter:
	"""
	Abstract class for filtering out samples matching some criteria. Initialise with data as needed,
	then filter out samples as needed.
	"""

	def filter_samples(self, samples: numpy.ndarray) -> numpy.ndarray:
		raise NotImplementedError


class DirectMonteCarloRun:
	"""
	A single model Direct Monte Carlo run, currently implemented only using single threading.
	An encapsulation of the steps needed for a Bayes run.

	Parameters
	----------
	model_name_pairs : Sequence[Tuple(str, pdme.model.DipoleModel)]
	The models to evaluate, with names

	measurements: Sequence[pdme.measurement.DotRangeMeasurement]
	The measurements as dot ranges to use as the bounds for the Monte Carlo calculation.

	monte_carlo_count_per_cycle: int
	The number of Monte Carlo iterations to use in a single cycle calculation.

	monte_carlo_cycles: int
	The number of cycles to use in each step.
	Increasing monte_carlo_count_per_cycle increases memory usage (and runtime), while this increases runtime, allowing
	control over memory use.

	target_success: int
	The number of successes to target before exiting early.
	Should likely be ~100 but can go higher to.

	max_monte_carlo_cycles_steps: int
	The number of steps to use. Each step consists of monte_carlo_cycles cycles, each of which has monte_carlo_count_per_cycle iterations.

	monte_carlo_seed: int
	The seed to use for the RNG.
	"""

	def __init__(
		self,
		model_name_pairs: Sequence[Tuple[str, pdme.model.DipoleModel]],
		filter: DirectMonteCarloFilter,
		config: DirectMonteCarloConfig,
	):
		self.model_name_pairs = model_name_pairs

		# self.measurements = measurements
		# self.dot_inputs = [(measure.r, measure.f) for measure in self.measurements]

		# self.dot_inputs_array = pdme.measurement.input_types.dot_inputs_to_array(
		# 	self.dot_inputs
		# )

		self.config = config
		self.filter = filter
		# (
		# 	self.lows,
		# 	self.highs,
		# ) = pdme.measurement.input_types.dot_range_measurements_low_high_arrays(
		# 	self.measurements
		# )

	def _single_run(
		self, model_name_pair: Tuple[str, pdme.model.DipoleModel], seed
	) -> numpy.ndarray:
		rng = numpy.random.default_rng(seed)

		_, model = model_name_pair
		# don't log here it's madness
		# _logger.info(f"Executing for model {model_name}")

		sample_dipoles = model.get_monte_carlo_dipole_inputs(
			self.config.monte_carlo_count_per_cycle, -1, rng
		)

		current_sample = sample_dipoles

		return self.filter.filter_samples(current_sample)
		# for di, low, high in zip(self.dot_inputs_array, self.lows, self.highs):

		# 	if len(current_sample) < 1:
		# 		break
		# 	vals = pdme.util.fast_v_calc.fast_vs_for_dipoleses(
		# 		numpy.array([di]), current_sample
		# 	)

		# 	current_sample = current_sample[
		# 		numpy.all((vals > low) & (vals < high), axis=1)
		# 	]
		# return current_sample

	def _wrapped_single_run(self, args: Tuple):
		"""
		single run wrapped up for multiprocessing call.

		takes in a tuple of arguments corresponding to
		(model_name_pair, seed)
		"""
		# here's where we do our work

		model_name_pair, seed = args
		cycle_success_configs = self._single_run(model_name_pair, seed)
		cycle_success_count = len(cycle_success_configs)

		return cycle_success_count

	def execute_no_multiprocessing(self) -> Sequence[DirectMonteCarloResult]:

		count_per_step = (
			self.config.monte_carlo_count_per_cycle * self.config.monte_carlo_cycles
		)
		seed_sequence = numpy.random.SeedSequence(self.config.monte_carlo_seed)

		# core count etc. logic here

		results = []
		for model_name_pair in self.model_name_pairs:
			step_count = 0
			total_success = 0
			total_count = 0

			_logger.info(f"Working on model {model_name_pair[0]}")
			# This is probably where multiprocessing logic should go
			while (step_count < self.config.max_monte_carlo_cycles_steps) and (
				total_success < self.config.target_success
			):
				_logger.debug(f"Executing step {step_count}")
				for cycle_i, seed in enumerate(
					seed_sequence.spawn(self.config.monte_carlo_cycles)
				):
					# here's where we do our work
					cycle_success_configs = self._single_run(model_name_pair, seed)
					cycle_success_count = len(cycle_success_configs)
					if cycle_success_count > 0:
						_logger.debug(
							f"For cycle {cycle_i} received {cycle_success_count} successes"
						)
						# _logger.debug(cycle_success_configs)
						if self.config.write_successes_to_file:
							sorted_by_freq = numpy.array(
								[
									pdme.subspace_simulation.sort_array_of_dipoles_by_frequency(
										dipole_config
									)
									for dipole_config in cycle_success_configs
								]
							)
							dipole_count = numpy.array(cycle_success_configs).shape[1]
							for n in range(dipole_count):
								numpy.savetxt(
									f"{self.config.tag}_{step_count}_{cycle_i}_dipole_{n}.csv",
									sorted_by_freq[:, n],
									delimiter=",",
								)
					total_success += cycle_success_count
				_logger.debug(
					f"At end of step {step_count} have {total_success} successes"
				)
				step_count += 1
				total_count += count_per_step

			results.append(
				DirectMonteCarloResult(
					successes=total_success,
					monte_carlo_count=total_count,
					likelihood=total_success / total_count,
					model_name=model_name_pair[0],
				)
			)
		return results

	def execute(self) -> Sequence[DirectMonteCarloResult]:

		# set starting execution timestamp
		timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")

		count_per_step = (
			self.config.monte_carlo_count_per_cycle * self.config.monte_carlo_cycles
		)
		seed_sequence = numpy.random.SeedSequence(self.config.monte_carlo_seed)

		# core count etc. logic here
		core_count = multiprocessing.cpu_count() - 1 or 1
		if (self.config.cap_core_count >= 1) and (
			self.config.cap_core_count < core_count
		):
			core_count = self.config.cap_core_count
		_logger.info(f"Using {core_count} cores")

		results = []
		with multiprocessing.Pool(core_count) as pool:

			for model_name_pair in self.model_name_pairs:
				_logger.info(f"Working on model {model_name_pair[0]}")
				# This is probably where multiprocessing logic should go

				step_count = 0
				total_success = 0
				total_count = 0

				while (step_count < self.config.max_monte_carlo_cycles_steps) and (
					total_success < self.config.target_success
				):

					step_count += 1

					_logger.debug(f"Executing step {step_count}")

					seeds = seed_sequence.spawn(self.config.monte_carlo_cycles)

					pool_results = sum(
						pool.imap_unordered(
							self._wrapped_single_run,
							[(model_name_pair, seed) for seed in seeds],
							self.config.chunk_size,
						)
					)
					_logger.debug(f"Pool results: {pool_results}")

					total_success += pool_results
					total_count += count_per_step
					_logger.debug(
						f"At end of step {step_count} have {total_success} successes"
					)

				results.append(
					DirectMonteCarloResult(
						successes=total_success,
						monte_carlo_count=total_count,
						likelihood=total_success / total_count,
						model_name=model_name_pair[0],
					)
				)

		if self.config.write_bayesrun_file:

			filename = (
				f"{timestamp}-{self.config.tag}.realdata.fast_filter.bayesrun.csv"
			)
			_logger.info(f"Going to write to file [{filename}]")
			# row: Dict[str, Union[int, float, str]] = {}
			row = {}

			num_models = len(self.model_name_pairs)
			success_weight = sum(
				[
					(res.successes / res.monte_carlo_count) / num_models
					for res in results
				]
			)

			for res in results:
				row.update(
					{
						f"{res.model_name}_success": res.successes,
						f"{res.model_name}_count": res.monte_carlo_count,
						f"{res.model_name}_prob": (
							res.successes / res.monte_carlo_count
						)
						/ (num_models * success_weight),
					}
				)
			_logger.info(f"Writing row {row}")
			fieldnames = list(row.keys())

			with open(filename, "w", newline="") as outfile:
				writer = csv.DictWriter(outfile, fieldnames=fieldnames, dialect="unix")
				writer.writeheader()
				writer.writerow(row)

		return results
